/**
* Zeigt durch sqlCodingTask definierte Level zum Lösen an
*/
function SqlCodingTaskUI(
    nodeContainer,      // DOM element to occupy
    levels              // Array describing levels
) {
    nodeContainer.innerHTML = '';
    this._msgPlaceGeneral = new ScoTools.MessagePlace();
    this._msgPlaceGeneral.appendAsChild(nodeContainer);
    this._levelMgr = new Levelliste(nodeContainer);
    this._levelMgr.onchange = ((levelNr, oldLevelNr) => this._startLevel(levelNr, oldLevelNr));
    this._levelMgr.strLevelName = 'Aufgabe ';
    this._levelMgr.onnamelevel = (nr => (nr > this.levels.length) ? 'Abschluss' : '');
    this._showTask('Wird geladen…');
    this._contentContainer = ScoTools.createHtml('div', nodeContainer);
    this._terminal = new SqlQueryTerminal();
    this._contentContainer.appendChild(this._terminal.node);
    this._terminal.onbeforeExecution = ((code, tokens, query) => this._procBeforeExecution(code, tokens, query));
    this._terminal.onafterExecution = ((results, code, tokens, query) => this._procAfterExecution(results, code, tokens, query));
    this._terminal.oncodeChange = (code => this._procCodeChange(code));
    this.loadLevels(levels);
}

// MessagePlace-Instanz zur Anzeige von grundlegenden Meldungen über die Speicherung des Lernerfolgs
SqlCodingTaskUI.prototype._msgPlaceGeneral = null;

// Verwendete Levelliste-Instanz
SqlCodingTaskUI.prototype._levelMgr = null;

// Verwendetes Eingabefeld für SQL-Abfrage
SqlCodingTaskUI.prototype._terminal = null;

// Container (DOM-Knoten) für Eingabefeld, ggf. Testfrage und Weiter-Schaltfläche
SqlCodingTaskUI.prototype._contentContainer = null;

// Container (DOM-Knoten) für Testfrage
SqlCodingTaskUI.prototype._containerQuestion = null;

// DOM-Knoten für „Weiter“-Schaltfläche
SqlCodingTaskUI.prototype._btnNext = null;

// Was muss als nächstes getan werden? (0 → Programmieraufgabe lösen, 1 → Frage richtig beantworten, 2 → Nichts mehr)
SqlCodingTaskUI.prototype.unlockNext = 0;

// Funktion, die bei Veränderung des Quellcodes (durch neues Level oder Benutzereingabe) aufgerufen werden soll
SqlCodingTaskUI.prototype.oncodeChange = function(code) {};

// Funktion, die bei Änderung des aktiven Levels aufgerufen werden soll
SqlCodingTaskUI.prototype.onlevelChange = function(index) {};

SqlCodingTaskUI.prototype._procCodeChange = function(code) {
    if (this.oncodeChange instanceof Function)
        this.oncodeChange.call(this, code);
}

// Funktion, die beim Freischalten eines neuen Levels aufgerufen werden soll
SqlCodingTaskUI.prototype.onlevelunlock = function(unlockedLevelCount) { }

// Zahl der bereits freigeschalteten Level
Object.defineProperty(SqlCodingTaskUI.prototype, 'unlockedLevelCount', {
    enumerable: true,
    get() { return Math.min(this._levelMgr.done, this.levels.length); },
    set(v) {
        this._levelMgr.done = v + ((this.levels.length == v) ? 1 : 0);
    }
});

// Wurden bereits alle Level frei geschaltet?
Object.defineProperty(SqlCodingTaskUI.prototype, 'done', {
    enumerable: true,
    get() { return (this._levelMgr.done > this.levels.length); }
});

// Funktion, die beim Erhalt von Strafpunkten aufgerufen werden soll
SqlCodingTaskUI.prototype.onnewfine = function(totalFine, additionalFine) { }

// Gesammelte Anzahl an Fehlerpunkten
SqlCodingTaskUI.prototype._fineCount = 0;
Object.defineProperty(SqlCodingTaskUI.prototype, 'fineCount', {
    enumerable: true,
    get() { return this._fineCount; },
    set(v) {
        if ((v == this._fineCount) || !Number.isFinite(v))
            return;
        // Fehlerpunkteanzahl nicht ändern, wenn ein bereits erledigtes Level erneut gespielt wird
        if (this._levelMgr.active < this.unlockedLevelCount)
            return;

        const oldFine = this._fineCount;
        this._fineCount = v;
        if ((typeof this.onnewfine == 'function') || (this.onnewfine instanceof Function))
            this.onnewfine.call(this, v, v - oldFine);
    }
});

// Eingegebene SQL-Abfrage
Object.defineProperty(SqlCodingTaskUI.prototype, 'userInput', {
    enumerable: true,
    get() { return this._terminal.code; },
    set(s) {
        if (this.userInput == s)
            return;
        const level = this.getLevelInfo();
        if (typeof level == 'object' && level.readonly)
            return;
        this._terminal.code = s;
    }
});

// Generierte Leveldaten
SqlCodingTaskUI.prototype._levels = [];
Object.defineProperty(SqlCodingTaskUI.prototype, 'levels', {
    enumerable: true,
    get() { return this._levels; },
    set(levels) {
        if (!(levels instanceof Array))
            return;
        this._levels = levels;
        this._levelMgr.count = levels.length + 1;
        if (this._levelMgr.active === undefined)
            this._levelMgr.active = 0;
    }
});

/**
* Load levels given by parameter
* @return undefined
*/
SqlCodingTaskUI.prototype.loadLevels = function(
    levels      // Levels (array or function returning array)
) {
    if (!(window.sqlCodingTask instanceof Object))
        return;
    if (levels instanceof Array) {
        this.levels = levels.slice();
        return;
    }
    if (levels instanceof Function) {
        try {
            this.levels = levels.call(levels);
        } catch (e) {}
    }
};

/**
* Aufgabenstellung anzeigen
* @return undefined
*/
SqlCodingTaskUI.prototype._showTask = function(
    task        // Anzuzeigender HTML-Code
) {
    if (!task && this._nodeTask) {
        this._nodeTask.parentNode.removeChild(this._nodeTask);
        this._nodeTask = null;
        return;
    }
    if (this._nodeTask) {
        this._nodeTask.innerHTML = task;
        return
    }
    this._nodeTask = ScoTools.createHtml('div', null, { 'class': 'auftrag' });
    this._nodeTask.innerHTML = task;
    this._levelMgr.node.parentNode.insertBefore(this._nodeTask, this._levelMgr.node.nextSibling);
}

/**
* Informationen über aktuelles Level ermitteln
* @return Object    Eintrag aus this.levels
*/
SqlCodingTaskUI.prototype.getLevelInfo = function() {
    const level = this.levels[this._levelMgr.active];
    return level ? level : null;
}

/** Wurde das aktuelle Level gelöst? */
SqlCodingTaskUI.prototype._levelSolved = false;

// Index des aktiven Levels
Object.defineProperty(SqlCodingTaskUI.prototype, 'levelIndex', {
    enumerable: true,
    get() { return this._levelMgr.active },
    set(index) { this._levelMgr.active = index; }
});

/**
* Level (neu) starten
* @return undefined
*/
SqlCodingTaskUI.prototype._startLevel = function(
    levelNr,        // Nummer des zu startenden Levels (beginnend bei 0)
    oldLevelNr      // Nummer des vorherigen Levels (beginnend bei 0)
) {
    if (this._changingLevel)
        return;
    this._changingLevel = true;
    const level = this.getLevelInfo();
    if (this.onlevelChange instanceof Function)
        this.onlevelChange.call(this, levelNr);
    if (level) {
        if (!this._contentContainer.parentNode) {
            this._levelMgr.node.parentNode.appendChild(this._contentContainer);
        }
        ['_containerQuestion', '_btnNext'].forEach(key => {
            if (this[key]) {
                if (this[key].parentNode)
                    this[key].parentNode.removeChild(this[key]);
                this[key] = null;
            }
        });

        const scope = new SqlScope();
        if (level.tables instanceof Array) {
            level.tables.forEach(({ table, name }) => {
                const reconstructedTable = SqlTable.fromJSON(table);
                if (reconstructedTable)
                    scope.addNamedTable(name, reconstructedTable);
            });
        }
        this._terminal.scope = scope;

        this.clearMsg();
        this._showTask(level.task ? level.task : '');
        if ((typeof level.start == 'function') || (level.start instanceof Function))
            level.start();
        if ((level.code !== undefined) && (!level.keepCode || !this._levelSolved || (oldLevelNr + 1 != levelNr)))
            this._terminal.code = level.code;
        this._terminal.editingEnabled = !level.readonly;
        this._terminal.hint = (level.hint === undefined) ? ('Verfügbare Tabellen: ' + scope.getTableSchemeHtml()) : level.hint;
        this.unlockNext = level.locked ? 0 : 1;
        this._levelSolved = false;
        this._terminal.clearMsg();
        this.showNextLock();
        this._procLevelStart();
    } else {
        if (this._contentContainer.parentNode)
            this._contentContainer.parentNode.removeChild(this._contentContainer);
        this._showTask(this.createFinalReport());
    }
    delete this._changingLevel;
}

/**
* Abschlussbericht zusammenstellen
* @return String        HTML-Code des Abschlussberichts
*/
SqlCodingTaskUI.prototype.createFinalReport = function() {
    if (!this.fineCount)
        return 'Hervorragend! Du hast alle Aufgaben fehlerfrei gelöst.';
    if (this.fineCount == 1)
        return 'Gut gemacht: Du hast alle Aufgaben gelöst und nur <strong>einen Fehlerpunkt</strong> kassiert.';
    const arrZahlen = ['', '', 'zwei', 'drei', 'vier', 'fünf', 'sechs', 'sieben', 'acht', 'neun', 'zehn', 'elf', 'zwölf'];
    if (this.fineCount < arrZahlen.length)
        return 'Du hast alle Aufgaben gemeistert, dabei allerdings <strong>' + arrZahlen[this.fineCount] + ' Fehlerpunkte</strong> gesammelt.';
    return 'Du hast alle Aufgaben gemeistert, dabei allerdings <strong>' + this.fineCount + ' Fehlerpunkte</strong> gesammelt.';
}

/**
* Dem Level ggf. mitteilen, dass es soeben (neu) gestartet wurde
* @return undefined
*/
SqlCodingTaskUI.prototype._procLevelStart = function() {
    const level = this.levels[this._levelMgr.active];
    let result = undefined;
    if ((level instanceof Object) && (level.onstart instanceof Function))
        result = level.onstart.call(level);
    if ((level instanceof Object) && (level.onstart instanceof Array))
        result = level.onstart.reduce((prev, checker) => (prev || ((checker instanceof Function) ? checker.call(level) : checker)), undefined);
}

/** Zwischenspeicher für tokens und compiled */
SqlCodingTaskUI.prototype._cacheLast = { tokens: [], compiled: null };

/**
* Levelspezifisch Programmcode vor der Ausführung prüfen
*/
SqlCodingTaskUI.prototype._procBeforeExecution = function(
    code,       // Value of this.code
    tokens,     // Array of all tokens representing code
    query       // SqlExpression instance representing parsed code
) {
    const params = {code, tokens, query, scope: this._terminal.scope, result: null};
    const level = this.getLevelInfo();
    if (!(level instanceof Object))
        return { accept: true };
    let checkResult = undefined;
    if ((level instanceof Object) && (level.checkCode instanceof Function))
        checkResult = level.checkCode.call(level, params);
    if (level.checkCode instanceof Array)
        checkResult = level.checkCode.reduce((prev, checker) => {
            if (prev !== undefined)
                return prev;
            if (checker instanceof SqlChecker)
                return checker.execute(params) ? checker : undefined;
            if (typeof checker == 'function')
                return checker.call(level, params);
            if (checker instanceof Object && checker.type)  {
                checker = SqlChecker.fromObject(checker);
                return (checker instanceof SqlChecker && checker.execute(params)) ? checker : undefined;
            }
            return checker;
        }, undefined);
    checkResult = (checkResult === undefined) ? { accept: true } : ((checkResult instanceof Object) ? checkResult : { accept: checkResult });
    if (checkResult.unlock && (this.unlockNext < 1)) {
        this.unlockNext = 1;
        this.showNextLock();
    }
    if (checkResult.fine)
        this.fineCount += Number.isFinite(checkResult.fine) ? checkResult.fine : 1;
    return checkResult;
}

/**
* Levelspezifisch Programmcode nach der Ausführung prüfen
*/
SqlCodingTaskUI.prototype._procAfterExecution = function(
    result,     // SqlTable instance representing results table
    code,       // Value of this.code executed
    tokens,     // Array of all tokens representing code
    query       // SqlExpression instance representing parsed code
) {
    const params = {result, code, tokens, query, scope: this._terminal.scope };
    const level = this.getLevelInfo();
    if (!(level instanceof Object))
        return { accept: true };
    let checkResult = undefined;
    if (typeof level.checkTermination == "function")
        checkResult = level.checkTermination.call(level, params);
    if (level.checkTermination instanceof Array)
        checkResult = level.checkTermination.reduce((prev, checker) => {
            if (prev !== undefined)
                return prev;
            if (checker instanceof SqlChecker)
                return checker.execute(params) ? checker : undefined;
            if (typeof checker == 'function')
                return checker.call(level, params);
            if (checker instanceof Object && checker.type)  {
                checker = SqlChecker.fromObject(checker);
                return (checker instanceof SqlChecker && checker.execute(params)) ? checker : undefined;
            }
            return checker;
        }, undefined);
    checkResult = (checkResult === undefined) ? { accept: true } : ((checkResult instanceof Object) ? checkResult : { accept: checkResult });
    if (checkResult.unlock && (this.unlockNext < 1)) {
        this.unlockNext = 1;
        this.showNextLock();
    }
    if (checkResult.fine)
        this.fineCount += Number.isFinite(checkResult.fine) ? checkResult.fine : 1;
    if (checkResult.accept === undefined)
        checkResult.accept = true;
    return checkResult;
}

/**
* Aktive Meldung löschen
* @return undefined
*/
SqlCodingTaskUI.prototype.clearMsg = function() {
    this._terminal.clearMsg();
}

/**
* Eine Meldung vorübergehend anzeigen
* @return undefined
*/
SqlCodingTaskUI.prototype.addMsg = function(
    text,       // Meldungstext
    type        // Optional: Meldungsart (Standardwert: Message.TYPE_FAILED)
) {
    this._terminal.addMsg(text, type);
}

/**
* Anzuzeigende Meldung über Speicherung des Lernerfolgs festsetzen
* @return undefined
*/
SqlCodingTaskUI.prototype.setStoringMsg = function(
    text,       // Anzuzeigender Meldungstext (oder kein bzw. leerer String für keine Meldung)
    type        // Optional: Meldungsart (Standardwert: ScoTools.MSGTYPE_WARN)
) {
    if ((typeof text != 'string' && !(text instanceof String)) || !text.length) {
        this._msgPlaceGeneral.clear();
        return;
    }
    this._msgPlaceGeneral.add(text, type || ScoTools.MSGTYPE_WARN);
}

/**
* Darauf reagieren, dass Teilaufgabe gelöst wurde
* @return undefined
*/
SqlCodingTaskUI.prototype.showNextLock = function() {
    if (this.unlockNext < 1)
        return;
    if (this.unlockNext < 2) {
        const level = this.getLevelInfo();
        if (!level || !level.question)
            this.unlockNext = 2;
        else if (!this._containerQuestion) {
            this._containerQuestion = ScoTools.createHtml('form', this._contentContainer, { method: 'post', action: location.href });
            const fragenui = new FragenUI(level.question, this._containerQuestion);
            const submitBtn = ScoTools.createHtml('button', this._containerQuestion, { type: 'submit', class: 'fullwidth' }, 'Antwort überprüfen');
            let needReload = false;
            this._containerQuestion.addEventListener('submit', ev => {
                ev.preventDefault();
                ev.stopPropagation();
                if (needReload) {
                    this._contentContainer.removeChild(this._containerQuestion);
                    this._containerQuestion = null;
                    this.showNextLock();
                    return;
                }
                const result = fragenui.evaluate();
                if (!result.correct) {
                    needReload = true;
                    this.addMsg('Frage falsch beantwortet');
                    submitBtn.textContent = 'Erneut versuchen';
                    if (level.questionWrongFine)
                        this.fineCount += Number.isFinite(level.questionWrongFine) ? level.questionWrongFine : 1;
                    return;
                }
                submitBtn.parentNode.removeChild(submitBtn);
                if (this.unlockNext < 2)
                    this.unlockNext = 2;
                this.addMsg('Aufgabe gelöst', ScoTools.Message.TYPE_SUCCESS);
                this.showNextLock();
            });
        }
    }
    if (this.unlockNext >= 2) {
        const countDone = this._levelMgr.active + ((this._levelMgr.active + 1 == this.levels.length) ? 2 : 1);
        if (countDone > this._levelMgr.done) {
            this._levelMgr.done = countDone;
            if ((typeof this.onlevelunlock == 'function') || (this.onlevelunlock instanceof Function))
                this.onlevelunlock.call(this, Math.min(countDone, this.levels.length));
        }
        if (!this._btnNext) {
            this._levelSolved = true;
            this._btnNext = ScoTools.createHtml('button', this._contentContainer, { 'class': 'fullwidth', type: 'button' }, 'Aufgabe abschließen');
            this._btnNext.addEventListener('click', ev => {
                ev.preventDefault();
                ev.stopPropagation();
                ++this._levelMgr.active;
            });
        }
    }
}
